@@ -1,4 +1,5 @@
 import logging
+import numbers
 import operator
 from functools import reduce
 
@@ -10,6 +11,8 @@ from openTSNE import nearest_neighbors
 from openTSNE import utils
 from openTSNE.utils import is_package_installed
 
+import warnings
+
 log = logging.getLogger(__name__)
 
 
@@ -78,7 +81,7 @@ class Affinities:
 
 
 class PerplexityBasedNN(Affinities):
-    """Compute affinities using nearest neighbors.
+    """Compute standard, Gaussian affinities using nearest neighbors.
 
     Please see the :ref:`parameter-guide` for more information.
 
@@ -93,10 +96,11 @@ class PerplexityBasedNN(Affinities):
 
     method: str
         Specifies the nearest neighbor method to use. Can be ``exact``, ``annoy``,
-        ``pynndescent``, ``hnsw``, ``approx``, or ``auto`` (default). ``approx`` uses Annoy
-        if the input data matrix is not a sparse object and if Annoy supports
-        the given metric. Otherwise it uses Pynndescent. ``auto`` uses exact
-        nearest neighbors for N<1000 and the same heuristic as ``approx`` for N>=1000.
+        ``pynndescent``, ``hnsw``, ``approx``, or ``auto`` (default). ``approx``
+        uses Annoy if the input data matrix is not a sparse object and if Annoy
+        supports the given metric. Otherwise, it uses Pynndescent. ``auto`` uses
+        exact nearest neighbors for N<1000 and the same heuristic as ``approx``
+        for N>=1000.
 
     metric: Union[str, Callable]
         The metric to be used to compute affinities between points in the
@@ -106,8 +110,9 @@ class PerplexityBasedNN(Affinities):
         Additional keyword arguments for the metric function.
 
     symmetrize: bool
-        Symmetrize affinity matrix. Standard t-SNE symmetrizes the interactions
-        but when embedding new data, symmetrization is not performed.
+        Symmetrize the affinity matrix. During standard t-SNE optimization, the
+        affinities are symmetrized. However, when embedding new data points into
+        existing embeddings, symmetrization is not performed.
 
     n_jobs: int
         The number of threads to use while running t-SNE. This follows the
@@ -167,8 +172,8 @@ class PerplexityBasedNN(Affinities):
             else:
                 _k_neighbors = k_neighbors
 
-            self.perplexity = self.check_perplexity(perplexity, _k_neighbors)
-            if _k_neighbors > int(3 * self.perplexity):
+            effective_perplexity = self.check_perplexity(perplexity, _k_neighbors)
+            if _k_neighbors > int(3 * effective_perplexity):
                 log.warning(
                     "The k_neighbors value is over 3 times larger than the perplexity value. "
                     "This may result in an unnecessary slowdown."
@@ -181,7 +186,7 @@ class PerplexityBasedNN(Affinities):
 
         else:
             self.knn_index = knn_index
-            self.perplexity = self.check_perplexity(perplexity, self.knn_index.k)
+            effective_perplexity = self.check_perplexity(perplexity, self.knn_index.k)
             log.info("KNN index provided. Ignoring KNN-related parameters.")
 
         self.__neighbors, self.__distances = self.knn_index.build()
@@ -190,11 +195,13 @@ class PerplexityBasedNN(Affinities):
             self.P = joint_probabilities_nn(
                 self.__neighbors,
                 self.__distances,
-                [self.perplexity],
+                [effective_perplexity],
                 symmetrize=symmetrize,
                 n_jobs=n_jobs,
             )
 
+        self.perplexity = perplexity
+        self.effective_perplexity_ = effective_perplexity
         self.symmetrize = symmetrize
         self.n_jobs = n_jobs
         self.verbose = verbose
@@ -220,27 +227,28 @@ class PerplexityBasedNN(Affinities):
         if new_perplexity == self.perplexity:
             return
         # Verify that the perplexity isn't negative
-        new_perplexity = self.check_perplexity(new_perplexity, np.inf)
+        effective_perplexity = self.check_perplexity(new_perplexity, np.inf)
         # Verify that the perplexity isn't too large for the kNN graph
-        if new_perplexity > self.__neighbors.shape[1]:
+        if effective_perplexity > self.__neighbors.shape[1]:
             raise RuntimeError(
                 "The desired perplexity `%.2f` is larger than the kNN graph "
                 "allows. This would need to recompute the nearest neighbors, "
                 "which is not efficient. Please create a new `%s` instance "
                 "with the increased perplexity."
-                % (new_perplexity, self.__class__.__name__)
+                % (effective_perplexity, self.__class__.__name__)
             )
         # Warn if the perplexity is larger than the heuristic
-        if 3 * new_perplexity > self.__neighbors.shape[1]:
+        if 3 * effective_perplexity > self.__neighbors.shape[1]:
             log.warning(
                 "The new perplexity is quite close to the computed number of "
                 "nearest neighbors. The results may be unexpected. Consider "
                 "creating a new `%s` instance with the increased perplexity."
-                % (self.__class__.__name__)
+                % self.__class__.__name__
             )
 
         # Recompute the affinity matrix
         self.perplexity = new_perplexity
+        self.effective_perplexity_ = effective_perplexity
         k_neighbors = int(3 * new_perplexity)
 
         with utils.Timer(
@@ -249,7 +257,7 @@ class PerplexityBasedNN(Affinities):
             self.P = joint_probabilities_nn(
                 self.__neighbors[:, :k_neighbors],
                 self.__distances[:, :k_neighbors],
-                [self.perplexity],
+                [self.effective_perplexity_],
                 symmetrize=self.symmetrize,
                 n_jobs=self.n_jobs,
             )
@@ -308,7 +316,7 @@ class PerplexityBasedNN(Affinities):
         else:
             _k_neighbors = k_neighbors
 
-        perplexity = self.check_perplexity(perplexity, _k_neighbors)
+        effective_perplexity = self.check_perplexity(perplexity, _k_neighbors)
 
         neighbors, distances = self.knn_index.query(data, _k_neighbors)
 
@@ -316,7 +324,7 @@ class PerplexityBasedNN(Affinities):
             P = joint_probabilities_nn(
                 neighbors,
                 distances,
-                [perplexity],
+                [effective_perplexity],
                 symmetrize=False,
                 normalization="point-wise",
                 n_reference_samples=self.n_samples,
@@ -328,7 +336,8 @@ class PerplexityBasedNN(Affinities):
 
         return P
 
-    def check_perplexity(self, perplexity, k_neighbors):
+    @staticmethod
+    def check_perplexity(perplexity, k_neighbors):
         if perplexity <= 0:
             raise ValueError("Perplexity must be >=0. %.2f given" % perplexity)
 
@@ -345,7 +354,7 @@ class PerplexityBasedNN(Affinities):
 def get_knn_index(
     data, method, k, metric, metric_params=None, n_jobs=1, random_state=None, verbose=False
 ):
-    # If we're dealing with a precomputed distance matrix, our job is very easy
+    # If we're dealing with a precomputed distance matrix, our job is very easy,
     # so we can skip all the remaining checks
     if metric == "precomputed":
         return nearest_neighbors.PrecomputedDistanceMatrix(data, k=k)
@@ -415,7 +424,7 @@ def joint_probabilities_nn(
     ----------
     neighbors: np.ndarray
         A `n_samples * k_neighbors` matrix containing the indices to each
-        points" nearest neighbors in descending order.
+        points' nearest neighbors in descending order.
     distances: np.ndarray
         A `n_samples * k_neighbors` matrix containing the distances to the
         neighbors at indices defined in the neighbors parameter.
@@ -486,7 +495,7 @@ def joint_probabilities_nn(
 
 
 class FixedSigmaNN(Affinities):
-    """Compute affinities using using nearest neighbors and a fixed bandwidth
+    """Compute affinities using nearest neighbors and a fixed bandwidth
     for the Gaussians in the ambient space.
 
     Using a fixed Gaussian bandwidth can enable us to find smaller clusters of
@@ -507,11 +516,11 @@ class FixedSigmaNN(Affinities):
 
     method: str
         Specifies the nearest neighbor method to use. Can be ``exact``, ``annoy``,
-        ``pynndescent``, ``hnsw``, ``approx``, or ``auto`` (default). ``approx`` uses Annoy
-        if the input data matrix is not a sparse object and if Annoy supports
-        the given metric. Otherwise it uses Pynndescent. ``auto`` uses exact
-        nearest neighbors for N<1000 and the same heuristic as ``approx`` for N>=1000.
-
+        ``pynndescent``, ``hnsw``, ``approx``, or ``auto`` (default). ``approx``
+        uses Annoy if the input data matrix is not a sparse object and if Annoy
+        supports the given metric. Otherwise, it uses Pynndescent. ``auto`` uses
+        exact nearest neighbors for N<1000 and the same heuristic as ``approx``
+        for N>=1000.
 
     metric: Union[str, Callable]
         The metric to be used to compute affinities between points in the
@@ -521,8 +530,9 @@ class FixedSigmaNN(Affinities):
         Additional keyword arguments for the metric function.
 
     symmetrize: bool
-        Symmetrize affinity matrix. Standard t-SNE symmetrizes the interactions
-        but when embedding new data, symmetrization is not performed.
+        Symmetrize the affinity matrix. During standard t-SNE optimization, the
+        affinities are symmetrized. However, when embedding new data points into
+        existing embeddings, symmetrization is not performed.
 
     n_jobs: int
         The number of threads to use while running t-SNE. This follows the
@@ -718,10 +728,11 @@ class MultiscaleMixture(Affinities):
 
     method: str
         Specifies the nearest neighbor method to use. Can be ``exact``, ``annoy``,
-        ``pynndescent``, ``hnsw``, ``approx``, or ``auto`` (default). ``approx`` uses Annoy
-        if the input data matrix is not a sparse object and if Annoy supports
-        the given metric. Otherwise it uses Pynndescent. ``auto`` uses exact
-        nearest neighbors for N<1000 and the same heuristic as ``approx`` for N>=1000.
+        ``pynndescent``, ``hnsw``, ``approx``, or ``auto`` (default). ``approx``
+        uses Annoy if the input data matrix is not a sparse object and if Annoy
+        supports the given metric. Otherwise, it uses Pynndescent. ``auto`` uses
+        exact nearest neighbors for N<1000 and the same heuristic as ``approx``
+        for N>=1000.
 
     metric: Union[str, Callable]
         The metric to be used to compute affinities between points in the
@@ -731,8 +742,9 @@ class MultiscaleMixture(Affinities):
         Additional keyword arguments for the metric function.
 
     symmetrize: bool
-        Symmetrize affinity matrix. Standard t-SNE symmetrizes the interactions
-        but when embedding new data, symmetrization is not performed.
+        Symmetrize the affinity matrix. During standard t-SNE optimization, the
+        affinities are symmetrized. However, when embedding new data points into
+        existing embeddings, symmetrization is not performed.
 
     n_jobs: int
         The number of threads to use while running t-SNE. This follows the
@@ -788,8 +800,8 @@ class MultiscaleMixture(Affinities):
             # We will compute the nearest neighbors to the max value of perplexity,
             # smaller values can just use indexing to truncate unneeded neighbors
             n_samples = data.shape[0]
-            perplexities = self.check_perplexities(perplexities, n_samples)
-            max_perplexity = np.max(perplexities)
+            effective_perplexities = self.check_perplexities(perplexities, n_samples)
+            max_perplexity = np.max(effective_perplexities)
             k_neighbors = min(n_samples - 1, int(3 * max_perplexity))
 
             self.knn_index = get_knn_index(
@@ -798,6 +810,8 @@ class MultiscaleMixture(Affinities):
 
         else:
             self.knn_index = knn_index
+            n_samples = self.knn_index.n_samples
+            effective_perplexities = self.check_perplexities(perplexities, n_samples)
             log.info("KNN index provided. Ignoring KNN-related parameters.")
 
         self.__neighbors, self.__distances = self.knn_index.build()
@@ -806,12 +820,13 @@ class MultiscaleMixture(Affinities):
             self.P = self._calculate_P(
                 self.__neighbors,
                 self.__distances,
-                perplexities,
+                effective_perplexities,
                 symmetrize=symmetrize,
                 n_jobs=n_jobs,
             )
 
         self.perplexities = perplexities
+        self.effective_perplexities_ = effective_perplexities
         self.symmetrize = symmetrize
         self.n_jobs = n_jobs
         self.verbose = verbose
@@ -856,8 +871,8 @@ class MultiscaleMixture(Affinities):
         if np.array_equal(self.perplexities, new_perplexities):
             return
 
-        new_perplexities = self.check_perplexities(new_perplexities, self.n_samples)
-        max_perplexity = np.max(new_perplexities)
+        effective_perplexities = self.check_perplexities(new_perplexities, self.n_samples)
+        max_perplexity = np.max(effective_perplexities)
         k_neighbors = min(self.n_samples - 1, int(3 * max_perplexity))
 
         if k_neighbors > self.__neighbors.shape[1]:
@@ -870,13 +885,14 @@ class MultiscaleMixture(Affinities):
             )
 
         self.perplexities = new_perplexities
+        self.effective_perplexities_ = effective_perplexities
         with utils.Timer(
             "Perplexity changed. Recomputing affinity matrix...", self.verbose
         ):
             self.P = self._calculate_P(
                 self.__neighbors[:, :k_neighbors],
                 self.__distances[:, :k_neighbors],
-                self.perplexities,
+                self.effective_perplexities_,
                 symmetrize=self.symmetrize,
                 n_jobs=self.n_jobs,
             )
@@ -923,9 +939,9 @@ class MultiscaleMixture(Affinities):
 
         """
         perplexities = perplexities if perplexities is not None else self.perplexities
-        perplexities = self.check_perplexities(perplexities, self.n_samples)
+        effective_perplexities = self.check_perplexities(perplexities, self.n_samples)
 
-        max_perplexity = np.max(perplexities)
+        max_perplexity = np.max(effective_perplexities)
         k_neighbors = min(self.n_samples - 1, int(3 * max_perplexity))
 
         neighbors, distances = self.knn_index.query(data, k_neighbors)
@@ -934,7 +950,7 @@ class MultiscaleMixture(Affinities):
             P = self._calculate_P(
                 neighbors,
                 distances,
-                perplexities,
+                effective_perplexities,
                 symmetrize=False,
                 normalization="point-wise",
                 n_reference_samples=self.n_samples,
@@ -954,6 +970,9 @@ class MultiscaleMixture(Affinities):
         value doesn't already exist in the list.
 
         """
+        if isinstance(perplexities, numbers.Number):
+            perplexities = [perplexities]
+
         usable_perplexities = []
         for perplexity in sorted(perplexities):
             if perplexity <= 0:
@@ -1002,10 +1021,11 @@ class Multiscale(MultiscaleMixture):
 
     method: str
         Specifies the nearest neighbor method to use. Can be ``exact``, ``annoy``,
-        ``pynndescent``, ``hnsw``, ``approx``, or ``auto`` (default). ``approx`` uses Annoy
-        if the input data matrix is not a sparse object and if Annoy supports
-        the given metric. Otherwise it uses Pynndescent. ``auto`` uses exact
-        nearest neighbors for N<1000 and the same heuristic as ``approx`` for N>=1000.
+        ``pynndescent``, ``hnsw``, ``approx``, or ``auto`` (default). ``approx``
+        uses Annoy if the input data matrix is not a sparse object and if Annoy
+        supports the given metric. Otherwise, it uses Pynndescent. ``auto`` uses
+        exact nearest neighbors for N<1000 and the same heuristic as ``approx``
+        for N>=1000.
 
     metric: Union[str, Callable]
         The metric to be used to compute affinities between points in the
@@ -1015,8 +1035,9 @@ class Multiscale(MultiscaleMixture):
         Additional keyword arguments for the metric function.
 
     symmetrize: bool
-        Symmetrize affinity matrix. Standard t-SNE symmetrizes the interactions
-        but when embedding new data, symmetrization is not performed.
+        Symmetrize the affinity matrix. During standard t-SNE optimization, the
+        affinities are symmetrized. However, when embedding new data points into
+        existing embeddings, symmetrization is not performed.
 
     n_jobs: int
         The number of threads to use while running t-SNE. This follows the
@@ -1074,7 +1095,7 @@ class Multiscale(MultiscaleMixture):
 
 
 class Uniform(Affinities):
-    """Compute affinities using using nearest neighbors and uniform kernel in
+    """Compute affinities using nearest neighbors and uniform kernel in
     the ambient space.
 
     Parameters
@@ -1086,11 +1107,11 @@ class Uniform(Affinities):
 
     method: str
         Specifies the nearest neighbor method to use. Can be ``exact``, ``annoy``,
-        ``pynndescent``, ``hnsw``, ``approx``, or ``auto`` (default). ``approx`` uses Annoy
-        if the input data matrix is not a sparse object and if Annoy supports
-        the given metric. Otherwise it uses Pynndescent. ``auto`` uses exact
-        nearest neighbors for N<1000 and the same heuristic as ``approx`` for N>=1000.
-
+        ``pynndescent``, ``hnsw``, ``approx``, or ``auto`` (default). ``approx``
+        uses Annoy if the input data matrix is not a sparse object and if Annoy
+        supports the given metric. Otherwise, it uses Pynndescent. ``auto`` uses
+        exact nearest neighbors for N<1000 and the same heuristic as ``approx``
+        for N>=1000.
 
     metric: Union[str, Callable]
         The metric to be used to compute affinities between points in the
@@ -1099,9 +1120,19 @@ class Uniform(Affinities):
     metric_params: dict
         Additional keyword arguments for the metric function.
 
-    symmetrize: bool
-        Symmetrize affinity matrix. Standard t-SNE symmetrizes the interactions
-        but when embedding new data, symmetrization is not performed.
+    symmetrize: Union[str, bool]
+        Symmetrize the affinity matrix. During standard t-SNE optimization, the
+        affinities are symmetrized. However, when embedding new data points into
+        existing embeddings, symmetrization is not performed.
+        The uniform affinity supports ``max`` and ``mean`` symmetrization, as
+        well as no symmetrization via ``none``.
+        The ``max`` symmetrization yields a binary affinity matrix with all
+        non-zero elements (corresponding to edges of the kNN graph) being the
+        same. The ``mean`` symmetrization performs symmetrization via
+        (A + A.T) / 2, resulting in the affinity matrix with two possible
+        non-zero values. Applying no symmetrization results in a non-symmetric
+        affinity matrix. We default to ``mean`` symmetrization, but the default
+        will change to ``max`` in future versions.
 
     n_jobs: int
         The number of threads to use while running t-SNE. This follows the
@@ -1176,8 +1207,24 @@ class Uniform(Affinities):
         )
 
         # Symmetrize the probability matrix
-        if symmetrize:
+        if symmetrize == "max":
+            P = (P + P.T > 0).astype(float)
+        elif symmetrize == "mean":
             P = (P + P.T) / 2
+        elif symmetrize == "none" or symmetrize is False:
+            pass
+        elif symmetrize is True:
+            # Backward compatibility
+            P = (P + P.T) / 2
+            warnings.warn(
+                f"Using `mean` symmetrization, but the default behaviour is going to "
+                f"change to `max` in future versions.",
+                category=FutureWarning,
+            )
+        else:
+            raise ValueError(
+                f"Symmetrization method `{symmetrize}` is not recognized."
+            )
 
         # Convert weights to probabilities
         P /= np.sum(P)
@@ -1252,3 +1299,26 @@ class Uniform(Affinities):
             return P, neighbors, distances
 
         return P
+
+
+class PrecomputedAffinities(Affinities):
+    """Use a precomputed affinity matrix.
+
+    Parameters
+    ----------
+    affinities: scipy.sparse.csr_matrix, np.ndarray
+        An N x N matrix containing the affinities.
+    normalize: bool
+        Normalize the affinity matrix to sum to 1. Default is True.
+
+    """
+
+    def __init__(self, affinities, normalize=True):
+        if not isinstance(affinities, sp.csr_matrix):
+            affinities = sp.csr_matrix(affinities)
+        if normalize:
+            affinities /= np.sum(affinities)
+        self.P = affinities
+
+    def to_new(self, data, return_distances=False):
+        raise RuntimeError("Precomputed affinity matrices cannot be queried.")
