@@ -10,7 +10,7 @@ from sklearn.base import BaseEstimator
 
 from openTSNE import _tsne
 from openTSNE import initialization as initialization_scheme
-from openTSNE.affinity import Affinities, PerplexityBasedNN
+from openTSNE.affinity import Affinities, MultiscaleMixture
 from openTSNE.quad_tree import QuadTree
 from openTSNE import utils
 
@@ -35,7 +35,7 @@ def _check_callbacks(callbacks):
 
 
 def _handle_nice_params(embedding: np.ndarray, optim_params: dict) -> None:
-    """Convert the user friendly params into something the optimizer can
+    """Convert the user-friendly params into something the optimizer can
     understand."""
     n_samples = embedding.shape[0]
     # Handle callbacks
@@ -43,8 +43,7 @@ def _handle_nice_params(embedding: np.ndarray, optim_params: dict) -> None:
     optim_params["use_callbacks"] = optim_params["callbacks"] is not None
 
     # Handle negative gradient method
-    negative_gradient_method = optim_params.pop("negative_gradient_method")
-    # Handle `auto` negative gradient method
+    negative_gradient_method = optim_params.pop("negative_gradient_method", "auto")
     if isinstance(negative_gradient_method, str) and negative_gradient_method == "auto":
         if n_samples < 10_000:
             negative_gradient_method = "bh"
@@ -88,8 +87,13 @@ def _handle_nice_params(embedding: np.ndarray, optim_params: dict) -> None:
     optim_params["n_jobs"] = n_jobs
 
     # Determine learning rate if requested
-    if optim_params.get("learning_rate", "auto") == "auto":
-        optim_params["learning_rate"] = max(200, n_samples / 12)
+    learning_rate = optim_params.get("learning_rate", "auto")
+    if learning_rate == "auto":
+        exaggeration = optim_params.get("exaggeration", None)
+        if exaggeration is None:
+            exaggeration = 1
+        learning_rate = n_samples / exaggeration
+    optim_params["learning_rate"] = learning_rate
 
 
 def __check_init_num_samples(num_samples, required_num_samples):
@@ -160,11 +164,12 @@ class PartialTSNEEmbedding(np.ndarray):
 
     learning_rate: Union[str, float]
         The learning rate for t-SNE optimization. When ``learning_rate="auto"``
-        the appropriate learning rate is selected according to max(200, N / 12)
+        the appropriate learning rate is selected according to N / exaggeration
         as determined in Belkina et al. (2019), Nature Communications. Note that
-        this should *not* be used when adding samples into existing embeddings,
-        where the learning rate often needs to be much lower to obtain
-        convergence.
+        this will result in a different learning rate during the early
+        exaggeration phase and afterwards. This should *not* be used when 
+        adding samples into existing embeddings, where the learning rate often
+        needs to be much lower to obtain convergence.
 
     exaggeration: float
         The exaggeration factor is used to increase the attractive forces of
@@ -242,9 +247,7 @@ class PartialTSNEEmbedding(np.ndarray):
     ):
         init_checks.num_samples(embedding.shape[0], P.shape[0])
 
-        obj = np.asarray(embedding, dtype=np.float64, order="C").view(
-            PartialTSNEEmbedding
-        )
+        obj = np.array(embedding, dtype=np.float64, order="C").view(PartialTSNEEmbedding)
 
         obj.reference_embedding = reference_embedding
         obj.P = P
@@ -280,12 +283,12 @@ class PartialTSNEEmbedding(np.ndarray):
         learning_rate: Union[str, float]
             The learning rate for t-SNE optimization. When
             ``learning_rate="auto"`` the appropriate learning rate is selected
-            according to max(200, N / 12), as determined in Belkina et al.
-            "Automated optimized parameters for t-distributed stochastic
-            neighbor embedding improve visualization and analysis of large
-            datasets", 2019. Note that this should *not* be used when adding
-            samples into existing embeddings, where the learning rate often
-            needs to be much lower to obtain convergence.
+            according to N / exaggeration as determined in Belkina et al.
+            (2019), Nature Communications. Note that this will result in a
+            different learning rate during the early exaggeration phase and
+            afterwards. This should *not* be used when adding samples into
+            existing embeddings, where the learning rate often needs to be much
+            lower to obtain convergence.
 
         exaggeration: float
             The exaggeration factor is used to increase the attractive forces of
@@ -414,10 +417,12 @@ class TSNEEmbedding(np.ndarray):
 
     learning_rate: Union[str, float]
         The learning rate for t-SNE optimization. When ``learning_rate="auto"``
-        the appropriate learning rate is selected according to max(200, N / 12),
-        as determined in Belkina et al. "Automated optimized parameters for
-        T-distributed stochastic neighbor embedding improve visualization and
-        analysis of large datasets", 2019.
+        the appropriate learning rate is selected according to N / exaggeration
+        as determined in Belkina et al. (2019), Nature Communications. Note that
+        this will result in a different learning rate during the early
+        exaggeration phase and afterwards. This should *not* be used when 
+        adding samples into existing embeddings, where the learning rate often
+        needs to be much lower to obtain convergence.
 
     exaggeration: float
         The exaggeration factor is used to increase the attractive forces of
@@ -504,7 +509,7 @@ class TSNEEmbedding(np.ndarray):
     ):
         init_checks.num_samples(embedding.shape[0], affinities.P.shape[0])
 
-        obj = np.asarray(embedding, dtype=np.float64, order="C").view(TSNEEmbedding)
+        obj = np.array(embedding, dtype=np.float64, order="C").view(TSNEEmbedding)
 
         obj.affinities = affinities  # type: Affinities
         obj.gradient_descent_params = gradient_descent_params  # type: dict
@@ -554,10 +559,12 @@ class TSNEEmbedding(np.ndarray):
         learning_rate: Union[str, float]
             The learning rate for t-SNE optimization. When
             ``learning_rate="auto"`` the appropriate learning rate is selected
-            according to max(200, N / 12), as determined in Belkina et al.
-            "Automated optimized parameters for t-distributed stochastic
-            neighbor embedding improve visualization and analysis of large
-            datasets", 2019.
+            according to N / exaggeration as determined in Belkina et al.
+            (2019), Nature Communications. Note that this will result in a
+            different learning rate during the early exaggeration phase and
+            afterwards. This should *not* be used when adding samples into
+            existing embeddings, where the learning rate often needs to be much
+            lower to obtain convergence.
 
         exaggeration: float
             The exaggeration factor is used to increase the attractive forces of
@@ -693,7 +700,7 @@ class TSNEEmbedding(np.ndarray):
         early_exaggeration_iter=0,
         exaggeration=1.5,
         n_iter=250,
-        initial_momentum=0.5,
+        initial_momentum=0.8,
         final_momentum=0.8,
         max_grad_norm=0.25,
         max_step_norm=None,
@@ -731,21 +738,21 @@ class TSNEEmbedding(np.ndarray):
             initial point positions.
 
         learning_rate: Union[str, float]
-            The learning rate for t-SNE optimization. When
+            The learning rate for t-SNE optimization. When 
             ``learning_rate="auto"`` the appropriate learning rate is selected
-            according to max(200, N / 12), as determined in Belkina et al.
-            "Automated optimized parameters for t-distributed stochastic
-            neighbor embedding improve visualization and analysis of large
-            datasets", 2019. Note that this should *not* be used when adding
-            samples into existing embeddings, where the learning rate often
-            needs to be much lower to obtain convergence.
+            according to N / exaggeration as determined in Belkina et al.
+            (2019), Nature Communications. Note that this will result in a
+            different learning rate during the early exaggeration phase and
+            afterwards. This should *not* be used when adding samples into
+            existing embeddings, where the learning rate often needs to be much
+            lower to obtain convergence.
 
         early_exaggeration_iter: int
             The number of iterations to run in the *early exaggeration* phase.
 
         early_exaggeration: float
             The exaggeration factor to use during the *early exaggeration* phase.
-            Typical values range from 12 to 32.
+            Typical values range from 4 to 32.
 
         n_iter: int
             The number of iterations to run in the normal optimization regime.
@@ -781,24 +788,26 @@ class TSNEEmbedding(np.ndarray):
 
         """
 
-        # We check if the affinity `to_new` methods takes the `perplexity`
-        # parameter and raise an informative error if not. This happes when the
-        # user uses a non-standard affinity class e.g. multiscale, then attempts
-        # to add points via `transform`. These classes take `perplexities` and
-        # fail
+        # Since the standard usage of t-SNE uses perplexities, check if the
+        # currently used affinity class supports either the `perplexity` or
+        # `perplexities` parameters.
         affinity_signature = inspect.signature(self.affinities.to_new)
-        if "perplexity" not in affinity_signature.parameters:
+        if "perplexity" in affinity_signature.parameters:
+            affinity_params = {"perplexity": perplexity}
+        elif "perplexities" in affinity_signature.parameters:
+            affinity_params = {"perplexities": perplexity}
+        else:
             raise TypeError(
-                "`transform` currently does not support non `%s` type affinity "
-                "classes. Please use `prepare_partial` and `optimize` to add "
-                "points to the embedding." % PerplexityBasedNN.__name__
+                f"`transform` currently does not support "
+                f"{self.affinities.__class__.__name__} affinities. Please use "
+                f"`prepare_partial` and `optimize` to add points to the embedding."
             )
 
         # Center the current embedding
         self -= (np.max(self, axis=0) + np.min(self, axis=0)) / 2
 
         embedding = self.prepare_partial(
-            X, perplexity=perplexity, initialization=initialization, k=k
+            X, initialization=initialization, k=k, **affinity_params
         )
 
         try:
@@ -861,6 +870,20 @@ class TSNEEmbedding(np.ndarray):
             optimization.
 
         """
+
+        # To maintain perfect backwards compatibility and to handle the very
+        # specific case when the user wants to pass in `perplexity` to the
+        # multiscale affinity object (multiscale accepts `perplexities`), rename
+        # this parameter so everything works
+        affinity_signature = inspect.signature(self.affinities.to_new)
+        if (
+            "perplexities" in affinity_signature.parameters and
+            "perplexities" in affinity_signature.parameters and
+            "perplexity" in affinity_params and
+            "perplexities" not in affinity_params
+        ):
+            affinity_params["perplexities"] = affinity_params.pop("perplexity")
+
         P, neighbors, distances = self.affinities.to_new(
             X, return_distances=True, **affinity_params
         )
@@ -979,17 +1002,22 @@ class TSNE(BaseEstimator):
 
     learning_rate: Union[str, float]
         The learning rate for t-SNE optimization. When ``learning_rate="auto"``
-        the appropriate learning rate is selected according to max(200, N / 12),
-        as determined in Belkina et al. "Automated optimized parameters for
-        T-distributed stochastic neighbor embedding improve visualization and
-        analysis of large datasets", 2019.
+        the appropriate learning rate is selected according to N / exaggeration
+        as determined in Belkina et al. (2019), Nature Communications. Note that
+        this will result in a different learning rate during the early
+        exaggeration phase and afterwards. This should *not* be used when 
+        adding samples into existing embeddings, where the learning rate often
+        needs to be much lower to obtain convergence.
 
     early_exaggeration_iter: int
         The number of iterations to run in the *early exaggeration* phase.
 
-    early_exaggeration: float
+    early_exaggeration: Union[str, float]
         The exaggeration factor to use during the *early exaggeration* phase.
-        Typical values range from 12 to 32.
+        Typical values range from 4 to 32. When ``early_exaggeration="auto"``
+        early exaggeration factor defaults to 12, unless desired subsequent
+        exaggeration is higher, i.e.: ``early_exaggeration = max(12,
+        exaggeration)``.
 
     n_iter: int
         The number of iterations to run in the normal optimization regime.
@@ -1106,7 +1134,7 @@ class TSNE(BaseEstimator):
         perplexity=30,
         learning_rate="auto",
         early_exaggeration_iter=250,
-        early_exaggeration=12,
+        early_exaggeration="auto",
         n_iter=500,
         exaggeration=None,
         dof=1,
@@ -1117,7 +1145,7 @@ class TSNE(BaseEstimator):
         initialization="pca",
         metric="euclidean",
         metric_params=None,
-        initial_momentum=0.5,
+        initial_momentum=0.8,
         final_momentum=0.8,
         max_grad_norm=None,
         max_step_norm=5,
@@ -1132,7 +1160,13 @@ class TSNE(BaseEstimator):
         self.n_components = n_components
         self.perplexity = perplexity
         self.learning_rate = learning_rate
-        self.early_exaggeration = early_exaggeration
+        if early_exaggeration == "auto":
+            if exaggeration is None:
+                self.early_exaggeration = 12
+            else:
+                self.early_exaggeration = max(12, exaggeration)
+        else:
+            self.early_exaggeration = early_exaggeration
         self.early_exaggeration_iter = early_exaggeration_iter
         self.n_iter = n_iter
         self.exaggeration = exaggeration
@@ -1185,7 +1219,7 @@ class TSNE(BaseEstimator):
 
         Parameters
         ----------
-        X: Optional[np.ndarray}
+        X: Optional[np.ndarray]
             The data matrix to be embedded.
 
         affinities: Optional[openTSNE.affinity.Affinities]
@@ -1255,7 +1289,7 @@ class TSNE(BaseEstimator):
 
         Parameters
         ----------
-        X: Optional[np.ndarray}
+        X: Optional[np.ndarray]
             The data matrix to be embedded.
 
         affinities: Optional[openTSNE.affinity.Affinities]
@@ -1287,7 +1321,7 @@ class TSNE(BaseEstimator):
         # If precomputed affinites are given, use those, otherwise proceed with
         # standard perpelxity-based affinites
         if affinities is None:
-            affinities = PerplexityBasedNN(
+            affinities = MultiscaleMixture(
                 X,
                 self.perplexity,
                 method=self.neighbors,
@@ -1538,11 +1572,14 @@ def kl_divergence_fft(
 class gradient_descent:
     def __init__(self):
         self.gains = None
+        self.update = None
 
     def copy(self):
         optimizer = self.__class__()
         if self.gains is not None:
             optimizer.gains = np.copy(self.gains)
+        if self.update is not None:
+            optimizer.update = np.copy(self.update)
         return optimizer
 
     def __call__(
@@ -1551,8 +1588,8 @@ class gradient_descent:
         P,
         n_iter,
         objective_function,
-        learning_rate=200,
-        momentum=0.5,
+        learning_rate="auto",
+        momentum=0.8,
         exaggeration=None,
         dof=1,
         min_gain=0.01,
@@ -1589,10 +1626,12 @@ class gradient_descent:
         learning_rate: Union[str, float]
             The learning rate for t-SNE optimization. When
             ``learning_rate="auto"`` the appropriate learning rate is selected
-            according to max(200, N / 12), as determined in Belkina et al.
-            "Automated optimized parameters for t-distributed stochastic
-            neighbor embedding improve visualization and analysis of large
-            datasets", 2019.
+            according to N / exaggeration as determined in Belkina et al.
+            (2019), Nature Communications. Note that this will result in a
+            different learning rate during the early exaggeration phase and
+            afterwards. This should *not* be used when adding samples into
+            existing embeddings, where the learning rate often needs to be much
+            lower to obtain convergence.
 
         momentum: float
             Momentum accounts for gradient directions from previous iterations,
@@ -1689,7 +1728,11 @@ class gradient_descent:
             )
 
         # If the interpolation grid has not yet been evaluated, do it now
-        if reference_embedding is not None and reference_embedding.interp_coeffs is None:
+        if (
+            reference_embedding is not None and
+            reference_embedding.interp_coeffs is None and
+            objective_function is kl_divergence_fft
+        ):
             reference_embedding.prepare_interpolation_grid()
 
         # If we're running transform and using the interpolation scheme, then we
@@ -1701,7 +1744,8 @@ class gradient_descent:
                 lower_limit = reference_embedding.box_x_lower_bounds[0]
                 upper_limit = reference_embedding.box_x_lower_bounds[-1]
 
-        update = np.zeros_like(embedding)
+        if self.update is None:
+            self.update = np.zeros_like(embedding).view(np.ndarray)
         if self.gains is None:
             self.gains = np.ones_like(embedding).view(np.ndarray)
 
@@ -1779,22 +1823,23 @@ class gradient_descent:
                     raise OptimizationInterrupt(error=error, final_embedding=embedding)
 
             # Update the embedding using the gradient
-            grad_direction_flipped = np.sign(update) != np.sign(gradient)
+            grad_direction_flipped = np.sign(self.update) != np.sign(gradient)
             grad_direction_same = np.invert(grad_direction_flipped)
             self.gains[grad_direction_flipped] += 0.2
             self.gains[grad_direction_same] = (
                 self.gains[grad_direction_same] * 0.8 + min_gain
             )
-            update = momentum * update - learning_rate * self.gains * gradient
+            gradient = gradient.view(np.ndarray)
+            self.update = momentum * self.update - learning_rate * self.gains * gradient
 
             # Clip the update sizes
             if max_step_norm is not None:
-                update_norms = np.linalg.norm(update, axis=1, keepdims=True)
+                update_norms = np.linalg.norm(self.update, axis=1, keepdims=True)
                 mask = update_norms.squeeze() > max_step_norm
-                update[mask] /= update_norms[mask]
-                update[mask] *= max_step_norm
+                self.update[mask] /= update_norms[mask]
+                self.update[mask] *= max_step_norm
 
-            embedding += update
+            embedding += self.update
 
             # Zero-mean the embedding only if we're not adding new data points,
             # otherwise this will reset point positions
