@@ -18,7 +18,9 @@ from openTSNE import affinity
 from openTSNE import initialization
 from openTSNE.affinity import PerplexityBasedNN
 from openTSNE.nearest_neighbors import NNDescent
-from openTSNE.tsne import kl_divergence_bh, kl_divergence_fft
+from openTSNE.tsne import (
+    kl_divergence_bh, kl_divergence_fft, TSNEEmbedding, PartialTSNEEmbedding
+)
 from openTSNE.utils import is_package_installed
 
 np.random.seed(42)
@@ -274,8 +276,9 @@ class TestTSNEInplaceOptimization(unittest.TestCase):
     @classmethod
     def setUpClass(cls):
         cls.tsne = TSNE()
-        cls.x = np.random.randn(100, 4)
-        cls.x_test = np.random.randn(25, 4)
+        cls.x = datasets.load_iris()["data"]
+        cls.x_test = cls.x[::3]
+        cls.x_test += np.random.normal(0, 1, size=cls.x_test.shape)
 
     def test_embedding_inplace_optimization(self):
         embedding1 = self.tsne.prepare_initial(self.x)
@@ -321,6 +324,38 @@ class TestTSNEInplaceOptimization(unittest.TestCase):
         self.assertFalse(partial_embedding2.base is partial_embedding3.base)
         self.assertFalse(partial_embedding1.base is partial_embedding3.base)
 
+    def test_inplace_embedding_optimization_doesnt_change_init(self):
+        init = initialization.pca(self.x)
+        init_copy = init.copy()
+
+        aff = affinity.PerplexityBasedNN(self.x)
+        embedding = TSNEEmbedding(init, aff)
+        embedding.optimize(10, inplace=True)
+
+        np.testing.assert_array_equal(init, init_copy)
+
+    def test_inplace_partial_embedding_optimization_doesnt_change_init(self):
+        embedding = TSNE(early_exaggeration_iter=10, n_iter=10).fit(self.x)
+
+        init = initialization.random(self.x_test)
+        init_copy = init.copy()
+
+        P = embedding.affinities.to_new(self.x_test)
+
+        partial_embedding = PartialTSNEEmbedding(init, embedding, P)
+        partial_embedding.optimize(10, inplace=True)
+
+        np.testing.assert_array_equal(init, init_copy)
+
+    def test_inplace_partial_embedding_optimization_doesnt_change_embedding(self):
+        embedding = TSNE(early_exaggeration_iter=10, n_iter=10).fit(self.x)
+        embedding_copy = np.array(embedding)
+
+        partial_embedding = embedding.prepare_partial(self.x_test)
+        partial_embedding.optimize(10, inplace=True)
+
+        np.testing.assert_array_equal(np.array(embedding), embedding_copy)
+
 
 class TestTSNECallbackParams(unittest.TestCase):
     @classmethod
@@ -431,9 +466,9 @@ class TestAlternativeFitUsageWithAffinityAndInitialization(unittest.TestCase):
 
     def test_pca_init_with_only_affinities_passed(self):
         aff = affinity.PerplexityBasedNN(self.x, 5, method="exact")
-        desired_init = initialization.spectral(aff.P)
+        desired_init = initialization.spectral(aff.P, random_state=42)
         embedding = TSNE(
-            early_exaggeration_iter=0, n_iter=0, initialization="pca"
+            early_exaggeration_iter=0, n_iter=0, initialization="pca", random_state=42
         ).fit(affinities=aff)
         np.testing.assert_array_equal(embedding, desired_init)
 
@@ -831,6 +866,8 @@ class TestGradientDescentOptimizer(unittest.TestCase):
             "Failed loading without any optimization",
         )
 
+        embedding = embedding.optimize(10)
+
         # After optimization
         loaded_embedding = pickle.loads(pickle.dumps(embedding))
         np.testing.assert_equal(
@@ -857,11 +894,20 @@ class TestAffinityIntegration(unittest.TestCase):
         # This should not raise an error
         embedding.transform(self.x_test)
 
+    def test_transform_with_multiscale_affinity(self):
+        init = openTSNE.initialization.random(self.x)
+        aff = openTSNE.affinity.Multiscale(self.x, [2, 5], method="exact")
+        embedding = openTSNE.TSNEEmbedding(init, aff, negative_gradient_method="bh")
+        embedding.optimize(100, inplace=True)
+
+        # This should not raise an error
+        embedding.transform(self.x_test)
+
     def test_transform_with_nonstandard_affinity(self):
         """Should raise an informative error when a non-standard affinity is used
         with `transform`."""
         init = openTSNE.initialization.random(self.x)
-        aff = openTSNE.affinity.Multiscale(self.x, [2, 5], method="exact")
+        aff = openTSNE.affinity.Uniform(self.x, 5, method="exact")
         embedding = openTSNE.TSNEEmbedding(init, aff, negative_gradient_method="bh")
         embedding.optimize(100, inplace=True)
 
@@ -891,8 +937,12 @@ class TestPrecomputedDistanceMatrices(unittest.TestCase):
         d = squareform(pdist(x))
 
         aff = affinity.PerplexityBasedNN(d, metric="precomputed")
-        desired_init = initialization.spectral(aff.P)
-        embedding = TSNE(early_exaggeration_iter=0, n_iter=0).fit(affinities=aff)
+        desired_init = initialization.spectral(aff.P, random_state=42)
+        embedding = TSNE(
+            early_exaggeration_iter=0, 
+            n_iter=0, 
+            random_state=42,
+        ).fit(affinities=aff)
         np.testing.assert_array_equal(embedding, desired_init)
 
     def test_precomputed_dist_matrix_via_tsne_interface_uses_spectral_init(self):
@@ -900,9 +950,13 @@ class TestPrecomputedDistanceMatrices(unittest.TestCase):
         d = squareform(pdist(x))
 
         aff = affinity.PerplexityBasedNN(d, metric="precomputed")
-        desired_init = initialization.spectral(aff.P)
-        embedding = TSNE(metric="precomputed", early_exaggeration_iter=0, n_iter=0) \
-            .fit(d)
+        desired_init = initialization.spectral(aff.P, random_state=42)
+        embedding = TSNE(
+            metric="precomputed", 
+            early_exaggeration_iter=0, 
+            n_iter=0,
+            random_state=42,
+         ).fit(d)
         np.testing.assert_array_equal(embedding, desired_init)
 
     def test_precomputed_dist_matrix_doesnt_override_valid_inits(self):
@@ -914,13 +968,13 @@ class TestPrecomputedDistanceMatrices(unittest.TestCase):
             initialization="random",
             metric="precomputed",
             early_exaggeration_iter=0,
-            n_iter=0
+            n_iter=0,
         ).fit(d)
 
         knn = KNeighborsClassifier(n_neighbors=10)
         knn.fit(embedding, y)
         predictions = knn.predict(embedding)
-        self.assertLess(accuracy_score(predictions, y), 0.55)
+        self.assertLess(accuracy_score(predictions, y), 0.6)
 
 
 class TestMisc(unittest.TestCase):
@@ -941,3 +995,14 @@ class TestMisc(unittest.TestCase):
 
         TSNE().fit(x, affinities=aff)
 
+    def test_interpolation_grid_not_called_using_bh(self):
+        x1 = np.random.normal(0, 1, (50, 10))
+        x2 = np.random.normal(0, 1, (20, 10))
+
+        with patch("openTSNE.TSNEEmbedding.prepare_interpolation_grid") as prep_grid:
+            tsne = openTSNE.TSNE(negative_gradient_method="bh")
+            embedding = tsne.fit(x1)
+            # Calling transform shouldn't call `prepare_interpolation_grid`
+            embedding.transform(x2)
+
+            prep_grid.assert_not_called()
